
#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/etherdevice.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <asm/tlbflush.h>

#include "lib/fc/fcconfig.h"
#include "lib/freqctrl.h"

#ifdef CONFIG_TANGO2
#define EM86XX_CHIP EM86XX_CHIPID_TANGO2
#include <linux/interrupt.h>
#include <asm/tango2/rmem86xxid.h>
#include <asm/tango2/rmdefs.h>
#include <asm/tango2/emhwlib_dram.h>
#include <asm/tango2/tango2_gbus.h>
#include <asm/tango2/tango2.h>
#include <asm/tango2/tango2api.h>
#include <asm/tango2/memcfg.h>
#elif defined(CONFIG_TANGO3)
#define EM86XX_CHIP EM86XX_CHIPID_TANGO3
#include <linux/interrupt.h>
#include <asm/tango3/rmem86xxid.h>
#include <asm/tango3/rmdefs.h>
#include <asm/tango3/emhwlib_dram.h>
#include <asm/tango3/tango3_gbus.h>
#include <asm/tango3/tango3.h>
#include <asm/tango3/tango3api.h>
#include <asm/tango3/memcfg.h>
#else
#error "Unknown architecture"
#endif

/* The names used in proc_fs */
#define PROC_ROOT               "tangoxfreq"
#define PROC_PROFILE            "profile"
#define PROC_MUX                "mux"
#define PROC_PLL3               "pll3"
#define PROC_PLL2               "pll2"
#define PROC_PLL1               "pll1"
#define PROC_PLL0               "pll0"
#define PROC_FREQ               "freq"

#ifdef WITH_INPUT_POLLING
#define PROC_STANDBY            "standby"
#ifdef WITH_IR
#define PROC_IR                 "ir_table"
#endif
#ifdef WITH_FIP
#define PROC_FIP                "fip_table"
#endif
#endif

/* Get XENV parameters related to frequency profile */
extern void tangox_get_ps_config(unsigned long *pll3, unsigned long *pll2, unsigned long *pll1, 
		unsigned long *pll0, unsigned long *mux, unsigned long *d0cfg, unsigned long *d0delay, 
		unsigned long *d1cfg, unsigned long *d1delay);

extern unsigned long tangox_get_cpuclock(void);
extern unsigned long tangox_get_sysclock(void);
extern unsigned long tangox_get_dspclock(void);
extern unsigned long tangox_get_pllclock(int idx);

static struct freq_profile freq_table[] = {
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "ORIGINAL" }, /* for saving the status when it's been loaded */
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, "PS XENV" }, /* for getting the ps status when it's been loaded from XENV */
	{ 0x0110009e, 0, 0, 0, 0x301, 0xf21111ba, 0x88888, 0xf21111ba, 0x88888, 0, "SYS/CPU/DSP 80/80/80" },  /* SYS/CPU/DSP: 80/80/80MHz, 2/2.5 */
	{ 0x011000c6, 0, 0, 0, 0x701, 0xf21111ba, 0x97777, 0xf21111ba, 0x97777, 0, "SYS/CPU/DSP 100/150/150" },  /* SYS/CPU/DSP: 150/100/150MHz, 2/2.5 */
	{ 0x01100108, 0, 0, 0, 0x701, 0xf21111ba, 0x97777, 0xf21111ba, 0x97777, 0, "SYS/CPU/DSP 133/200/200" },  /* SYS/CPU/DSP: 200/133/200MHz, 2/2.5 */
	{ 0x0110014a, 0, 0, 0, 0x701, 0xf34111ba, 0xa7777, 0xf34111ba, 0xa7777, 0, "SYS/CPU/DSP 166/250/250" },  /* SYS/CPU/DSP: 250/166/250MHz, 3/4 */
	{ 0x01010041, 0, 0, 0, 0x701, 0xf34111ba, 0xa4444, 0xf34111ba, 0xa4444, 0, "SYS/CPU/DSP 200/300/300" },  /* SYS/CPU/DSP: 300/200/300MHz, 3/4 */
};

#ifdef WITH_INPUT_POLLING
static const struct freq_profile standby_mode =
//      { 0x01010010, 0x202, 0x202, 0x202, 0x301, 0xf21111ba, 0x88888, 0xf21111ba, 0x88888, 1, "54/54/54" };  /* CPU/SYS/DSP: 54/54/54MHz, 2/2.5 PLL3 162MHz */
	{ 0x0119006a, 0x202, 0x202, 0x202, 0x1, 0xf21111ba, 0x88888, 0xf21111ba, 0x88888, 1, "54/54/54" };  /* CPU/SYS/DSP: 54/54/54MHz, 2/2.5 PLL3 108MHz */
#endif

static int profile_idx = 0;
static struct proc_dir_entry *root = NULL, *prof_ent = NULL, *mux_ent = NULL, *pll3_ent = NULL;
static struct proc_dir_entry *pll1_ent = NULL, *pll2_ent = NULL, *pll0_ent = NULL, *freq_ent = NULL;

#ifdef WITH_INPUT_POLLING
static struct proc_dir_entry *standby_ent;
#ifdef WITH_IR
static struct proc_dir_entry *ir_ent;
#endif
#ifdef WITH_FIP
static struct proc_dir_entry *fip_ent;
#endif
static unsigned long ir_keys[INPUT_TABLE_SIZE/sizeof(unsigned long)];
static unsigned long fip_keys[INPUT_TABLE_SIZE/sizeof(unsigned long)];
#endif

static const unsigned char ratios_table[12][3] = {
	{ 2, 2, 2 },
	{ 4, 2, 4 },
	{ 3, 2, 3 },
	{ 3, 3, 3 },
	{ 3, 3, 3 },
	{ 3, 2, 3 },
	{ 3, 3, 2 },
	{ 3, 2, 2 },
	{ 4, 4, 4 },
	{ 4, 2, 4 },
	{ 4, 4, 2 },
	{ 4, 2, 2 }
};

static int change_profile(const char *str)
{
	int prof = simple_strtol(str, NULL, 10);

	if ((prof < 0) || (prof >= (sizeof(freq_table)/sizeof(struct freq_profile)))) 
		return -EINVAL;

	for (; prof < sizeof(freq_table)/sizeof(struct freq_profile); prof++) {
		/* Look for defined profile */
		if (freq_table[prof].pll3 != 0)
			break;
	}

	/* No valid profile found, use original one */
	if (prof >= (sizeof(freq_table)/sizeof(struct freq_profile)))
		prof = 0; 

	if (profile_idx == prof)
		return 0;

	profile_idx = prof;

#ifdef WITH_INPUT_POLLING
	return set_freq_profile(&freq_table[prof], NULL, NULL, NULL, NULL);
#else
	return set_freq_profile(&freq_table[prof]);
#endif
}

static int change_mux(const char *mux_str)
{
	struct freq_profile freq_prof;
	int mux = simple_strtol(mux_str, NULL, 10);

	if ((mux < 0) || (mux > 11))
		return -EINVAL;

	freq_prof.mux = (gbus_readl(REG_BASE_system_block+SYS_sysclk_mux) & 0xfffff0ff) | (mux << 8);
	freq_prof.d0cfg = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_cfg);
	freq_prof.d0delays = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_delay0_ctrl);
	freq_prof.d1cfg = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_cfg);
	freq_prof.d1delays = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_delay0_ctrl);
	freq_prof.pll3 = gbus_readl(REG_BASE_system_block+SYS_clkgen3_pll);
	freq_prof.pll2 = freq_prof.pll1 = freq_prof.pll0 = 0;
	freq_prof.standby = 0;

	if (freq_prof.mux != gbus_readl(REG_BASE_system_block+SYS_sysclk_mux))
		profile_idx = -1;

#ifdef WITH_INPUT_POLLING
	return set_freq_profile(&freq_prof, NULL, NULL, NULL, NULL);
#else
	return set_freq_profile(&freq_prof);
#endif
}

static int pllx_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char buf[64];
	unsigned long freq;
	struct freq_profile freq_prof;
	int idx = (int)data;

	if (count > ARRAY_SIZE(buf) - 1)
		count = ARRAY_SIZE(buf) - 1;
	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';
	freq = simple_strtol(buf, NULL, 10);

	freq_prof.mux = gbus_readl(REG_BASE_system_block+SYS_sysclk_mux);
	freq_prof.d0cfg = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_cfg);
	freq_prof.d0delays = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_delay0_ctrl);
	freq_prof.d1cfg = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_cfg);
	freq_prof.d1delays = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_delay0_ctrl);
	freq_prof.standby = 0;
	freq_prof.pll3 = gbus_readl(REG_BASE_system_block+SYS_clkgen3_pll);
	freq_prof.pll2 = gbus_readl(REG_BASE_system_block+SYS_clkgen2_pll);
	freq_prof.pll1 = gbus_readl(REG_BASE_system_block+SYS_clkgen1_pll);
	freq_prof.pll0 = gbus_readl(REG_BASE_system_block+SYS_clkgen0_pll);

	if (idx == 3) {
		freq_prof.pll3 = frequency_to_pll(freq, 1);
	} else if (idx == 2) {
		freq_prof.pll2 = frequency_to_pll(freq, 1);
	} else if (idx == 1) {
		freq_prof.pll1 = frequency_to_pll(freq, 1);
	} else if (idx == 0) {
		freq_prof.pll0 = frequency_to_pll(freq, 1);
	}

	profile_idx = -1;
#ifdef WITH_INPUT_POLLING
	set_freq_profile(&freq_prof, NULL, NULL, NULL, NULL);
#else
	set_freq_profile(&freq_prof);
#endif

	return count;
}

static int pllx_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int ret = 0;
	int idx = (int)data;

	ret = sprintf(page + off, "PLL%d: %ldHz\n", idx, tangox_get_pllclock(idx));
	*eof = 1;
	return ret;
}

static int freq_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int ret = 0;
	int i;
	for (i = 0; i < 4; i++) {
		ret += sprintf(page + off + ret, "PLL%d: %ldHz\n", i, tangox_get_pllclock(i));
	}
	ret += sprintf(page + off + ret, "SYS: %ldHz\n", tangox_get_sysclock());
	ret += sprintf(page + off + ret, "CPU: %ldHz\n", tangox_get_cpuclock());
	ret += sprintf(page + off + ret, "DSP: %ldHz\n", tangox_get_dspclock());
	*eof = 1;
	return ret;
}

static int profile_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char buf[16];

	if (count > ARRAY_SIZE(buf) - 1)
		count = ARRAY_SIZE(buf) - 1;
	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';

	change_profile(buf);

	return count;
}

static int profile_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int ret = 0;
	int i;

	for (i = 0; i < (sizeof(freq_table)/sizeof(struct freq_profile)); i++) {
		if (freq_table[i].pll3 == 0)
			continue;
		ret += sprintf(page + off + ret, "%c %d: %s\n", (profile_idx == i) ? '*' : ' ', 
				i, freq_table[i].desc);
	}
	*eof = 1;
	return ret;
}

static int mux_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char buf[16];

	if (count > ARRAY_SIZE(buf) - 1)
		count = ARRAY_SIZE(buf) - 1;
	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';

	change_mux(buf);

	return count;
}

static int mux_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int ret = 0, i;
	unsigned long mux = (gbus_readl(REG_BASE_system_block+SYS_sysclk_mux) >> 8) & 0xf;
	unsigned long premux = gbus_readl(REG_BASE_system_block+SYS_sysclk_premux);
	
	ret = sprintf(page + off, "PLL%ld: %ldHz\n", premux & 3, tangox_get_pllclock(premux & 3));
	for (i = 0; i < 12; i++) {
		ret += sprintf(page + off + ret, "%c %d (SYS/CPU/DSP: %d/%d/%d)\n",
				(i == mux) ? '*' : ' ', i,
				ratios_table[i][0], ratios_table[i][1], ratios_table[i][2]);
	}
	*eof = 1;
	return ret;
}

#ifdef WITH_INPUT_POLLING
static int standby_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	char buf[16];
	unsigned long elapse, ret_key;
	unsigned long *ir_table, *fip_table;

	if (count > ARRAY_SIZE(buf) - 1)
		count = ARRAY_SIZE(buf) - 1;
	if (copy_from_user(buf, buffer, count))
		return -EFAULT;

	buf[count] = '\0';

#ifdef WITH_IR
	ir_table = ir_keys;
#else
	ir_table = NULL;
#endif
#ifdef WITH_FIP
	fip_table = fip_keys;
#else
	fip_table = NULL;
#endif

	if (strncmp("standby", buf, strlen("standby")) == 0) {
		printk("Entering standby mode ...\n");
		set_freq_profile(&standby_mode, &elapse, &ret_key, ir_table, fip_table);
		jiffies += ((elapse * HZ) / 1000); /* elaspe in msec so we adjust it accordingly */ 
		printk("Returned key 0x%08lx, time elapse %ld jiffies.\n", ret_key, (elapse * HZ) / 1000);
	}

	return count;
}
#ifdef WITH_IR
static int ir_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	return(count);
}
static int ir_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	*eof = 1;
	return(0);
}
#endif
#ifdef WITH_FIP
static int fip_info_write_proc(struct file *file, const char __user *buffer, unsigned long count, void *data)
{
	return(count);
}
static int fip_info_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	*eof = 1;
	return(0);
}
#endif
#endif

static int tangox_freq_ctrl_init(void) 
{
#ifdef WITH_INPUT_POLLING
	memset(ir_keys, 0, INPUT_TABLE_SIZE);
	memset(fip_keys, 0, INPUT_TABLE_SIZE);
#endif

	freq_table[0].pll3 = gbus_readl(REG_BASE_system_block+SYS_clkgen3_pll);
	freq_table[0].pll2 = gbus_readl(REG_BASE_system_block+SYS_clkgen2_pll);
	freq_table[0].pll1 = gbus_readl(REG_BASE_system_block+SYS_clkgen1_pll);
	freq_table[0].pll0 = gbus_readl(REG_BASE_system_block+SYS_clkgen0_pll);
	freq_table[0].mux = gbus_readl(REG_BASE_system_block+SYS_sysclk_mux);
	freq_table[0].d0cfg = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_cfg);
	freq_table[0].d0delays = gbus_readl(REG_BASE_dram_controller_0+DRAM_dunit_delay0_ctrl);
	freq_table[0].d1cfg = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_cfg);
	freq_table[0].d1delays = gbus_readl(REG_BASE_dram_controller_1+DRAM_dunit_delay0_ctrl);

	tangox_get_ps_config(&freq_table[1].pll3, &freq_table[1].pll2, &freq_table[1].pll1, 
			&freq_table[1].pll0, &freq_table[1].mux, &freq_table[1].d0cfg, 
			&freq_table[1].d0delays, &freq_table[1].d1cfg, &freq_table[1].d1delays);

	if ((root = proc_mkdir(PROC_ROOT, NULL)) == NULL)
		return -EINVAL;

	if ((prof_ent = create_proc_entry(PROC_PROFILE, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		prof_ent->read_proc = profile_info_read_proc;
		prof_ent->write_proc = profile_info_write_proc;
	}

	if ((mux_ent = create_proc_entry(PROC_MUX, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		mux_ent->read_proc = mux_info_read_proc;
		mux_ent->write_proc = mux_info_write_proc;
	}

	if ((freq_ent = create_proc_entry(PROC_FREQ, S_IRUGO, root)) == NULL) 
		return -EINVAL;
	else {
		freq_ent->read_proc = freq_info_read_proc;
	}

	if ((pll3_ent = create_proc_entry(PROC_PLL3, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		pll3_ent->read_proc = pllx_info_read_proc;
		pll3_ent->write_proc = pllx_info_write_proc;
		pll3_ent->data = (void *)3;
	}

	if ((pll2_ent = create_proc_entry(PROC_PLL2, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		pll2_ent->read_proc = pllx_info_read_proc;
		pll2_ent->write_proc = pllx_info_write_proc;
		pll2_ent->data = (void *)2;
	}

	if ((pll1_ent = create_proc_entry(PROC_PLL1, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		pll1_ent->read_proc = pllx_info_read_proc;
		pll1_ent->write_proc = pllx_info_write_proc;
		pll1_ent->data = (void *)1;
	}

	if ((pll0_ent = create_proc_entry(PROC_PLL0, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		pll0_ent->read_proc = pllx_info_read_proc;
		pll0_ent->write_proc = pllx_info_write_proc;
		pll0_ent->data = (void *)0;
	}

#ifdef WITH_INPUT_POLLING
	if ((standby_ent = create_proc_entry(PROC_STANDBY, S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else 
		standby_ent->write_proc = standby_info_write_proc;
#ifdef WITH_IR
	if ((ir_ent = create_proc_entry(PROC_IR, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		ir_ent->write_proc = ir_info_write_proc;
		ir_ent->read_proc = ir_info_read_proc;
	}
#endif
#ifdef WITH_FIP
	if ((fip_ent = create_proc_entry(PROC_FIP, S_IRUGO | S_IWUGO, root)) == NULL) 
		return -EINVAL;
	else {
		fip_ent->write_proc = fip_info_write_proc;
		fip_ent->read_proc = fip_info_read_proc;
	}
#endif
#endif

	return 0;
}

static void tangox_freq_ctrl_exit(void) /* Removes module from kernel */
{
	if (root) {
		if (prof_ent)
			remove_proc_entry(PROC_PROFILE, root);
		if (mux_ent)
			remove_proc_entry(PROC_MUX, root);
		if (pll3_ent)
			remove_proc_entry(PROC_PLL3, root);
		if (pll2_ent)
			remove_proc_entry(PROC_PLL2, root);
		if (pll1_ent)
			remove_proc_entry(PROC_PLL1, root);
		if (pll0_ent)
			remove_proc_entry(PROC_PLL0, root);
		if (freq_ent)
			remove_proc_entry(PROC_FREQ, root);
#ifdef WITH_INPUT_POLLING
		if (standby_ent)
			remove_proc_entry(PROC_STANDBY, root);
#ifdef WITH_IR
		if (ir_ent)
			remove_proc_entry(PROC_IR, root);
#endif
#ifdef WITH_FIP
		if (fip_ent)
			remove_proc_entry(PROC_FIP, root);
#endif
#endif
		remove_proc_entry(PROC_ROOT, NULL);
	}
}

module_init(tangox_freq_ctrl_init);
module_exit(tangox_freq_ctrl_exit);

MODULE_DESCRIPTION("SMP863x/SMP864x frequency scaling driver");
MODULE_AUTHOR("Craig");
MODULE_LICENSE("GPL");

