#############################################################
#
# cprops
#
#############################################################
CPROPS_VER=0.0.7
CPROPS_SOURCE=libcprops-$(CPROPS_VER).tar.gz
CPROPS_SITE=http://heanet.dl.sourceforge.net/sourceforge/cprops/
CPROPS_DIR=$(BUILD_DIR)/libcprops-$(CPROPS_VER)
CPROPS_CFLAGS= $(TARGET_CFLAGS)
CPROPS_TARGET_BINARY:=usr/lib/libcprops.so.$(CPROPS_VER)
CPROPS_BUILT_LIB=$(CPROPS_DIR)/.libs/libcprops.so.$(CPROPS_VER)
CPROPS_LIB_FOR_LINK=$(STAGING_DIR)/lib/libcprops.so.$(CPROPS_VER)
ifeq ($(BR2_LARGEFILE),y)
CPROPS_CFLAGS+= -D_LARGEFILE_SOURCE -D_LARGEFILE64_SOURCE -D_FILE_OFFSET_BITS=64
endif

$(DL_DIR)/$(CPROPS_SOURCE):
	$(WGET) -P $(DL_DIR) $(CPROPS_SITE)/$(CPROPS_SOURCE)

$(CPROPS_DIR)/.source: $(DL_DIR)/$(CPROPS_SOURCE) package/cprops/cprops.patch
	zcat $(DL_DIR)/$(CPROPS_SOURCE) | tar -C $(BUILD_DIR) $(TAR_OPTIONS) -
	cp package/cprops/cprops.patch $(CPROPS_DIR)
	(cd $(CPROPS_DIR); patch -p1 < cprops.patch)
	touch $(CPROPS_DIR)/.source

$(CPROPS_DIR)/.configured: $(CPROPS_DIR)/.source package/cprops/cprops.mk
	(cd $(CPROPS_DIR); \
		BUILD_CC=$(TARGET_CC) HOSTCC=$(HOSTCC) \
		$(TARGET_CONFIGURE_OPTS) \
		CFLAGS="$(TARGET_CFLAGS)" \
		./configure \
		--target=$(GNU_TARGET_NAME) \
		--host=$(GNU_TARGET_NAME) \
		--build=$(GNU_HOST_NAME) \
		--prefix=/usr \
		--libdir=$(STAGING_DIR)/lib \
		--includedir=$(STAGING_DIR)/include \
		--mandir=$(STAGING_DIR)/man \
		--disable-ssl \
		--disable-server-cookies \
		--disable-http-sessions \
	);
	touch $(CPROPS_DIR)/.configured;

$(CPROPS_BUILT_LIB): $(CPROPS_DIR)/.configured
	$(MAKE) BUILD_CC=$(TARGET_CC) HOSTCC=$(HOSTCC) \
		BUILD_LDFLAGS="" DESTDIR=$(STAGING_DIR) -C $(CPROPS_DIR)
	@touch -c $(CPROPS_BUILT_LIB)

$(CPROPS_LIB_FOR_LINK): $(CPROPS_BUILT_LIB)
	$(MAKE) -C $(CPROPS_DIR) install

$(TARGET_DIR)/$(CPROPS_TARGET_BINARY): $(CPROPS_LIB_FOR_LINK)
	cp -dpf $(STAGING_DIR)/lib/libcprops.so* $(TARGET_DIR)/usr/lib/
	@touch -c $(TARGET_DIR)/$(CPROPS_TARGET_BINARY)

cprops: $(TARGET_DIR)/$(CPROPS_TARGET_BINARY)

cprops-source: $(DL_DIR)/$(CPROPS_SOURCE)

cprops-clean:
	rm -f $(STAGING_DIR)/lib/libcprops.a
	rm -f $(STAGING_DIR)/lib/libcprops.so*
	rm -f $(TARGET_DIR)/$(CPROPS_TARGET_BINARY)
	-$(MAKE) -C $(CPROPS_DIR) clean
	rm -f $(CPROPS_DIR)/.configured;

cprops-dirclean:
	rm -rf $(CPROPS_DIR)

#############################################################
#
# Toplevel Makefile options
#
#############################################################
ifeq ($(strip $(BR2_PACKAGE_CPROPS)),y)
TARGETS+=cprops
endif
