#############################################################
#
# iptables
#
#############################################################

IPTABLES_DIR := package/iptables/source
IPTABLES_VER:=1.3.5
IPTABLES_SOURCE_URL:=ftp.netfilter.org/pub/iptables/
IPTABLES_SOURCE:=iptables-$(IPTABLES_VER).tar.bz2
IPTABLES_BUILD_DIR:=$(BUILD_DIR)/iptables-$(IPTABLES_VER)
IPTABLES_INIT_SCRIPT := firewall
IPTABLES_TARGET_INIT_SCRIPT := etc/init.d/firewall
IPTABLES_TARGET_RC_LINK := etc/rc.d/S15firewall

$(DL_DIR)/$(IPTABLES_SOURCE):
	 $(WGET) -P $(DL_DIR) $(IPTABLES_SOURCE_URL)/$(IPTABLES_SOURCE) 

$(IPTABLES_BUILD_DIR)/.unpacked: $(DL_DIR)/$(IPTABLES_SOURCE)
	bzcat $(DL_DIR)/$(IPTABLES_SOURCE) | tar -C $(BUILD_DIR) $(TAR_OPTIONS) -
	touch $(IPTABLES_BUILD_DIR)/.unpacked

$(IPTABLES_BUILD_DIR)/.configured: $(IPTABLES_BUILD_DIR)/.unpacked
	# Allow patches.  Needed for openwrt for instance.
	toolchain/patch-kernel.sh $(IPTABLES_BUILD_DIR) package/iptables/ iptables\*.patch
	#
	$(SED) "s;\[ -f /usr/include/netinet/ip6.h \];grep -q '__UCLIBC_HAS_IPV6__ 1' \
		$(STAGING_DIR)/include/bits/uClibc_config.h;" $(IPTABLES_BUILD_DIR)/Makefile
	touch $(IPTABLES_BUILD_DIR)/.configured

$(IPTABLES_BUILD_DIR)/iptables: $(IPTABLES_BUILD_DIR)/.configured
	$(TARGET_CONFIGURE_OPTS) \
	$(MAKE) -C $(IPTABLES_BUILD_DIR) \
		KERNEL_DIR=$(LINUX_DIR) PREFIX=/usr \
		CC=$(TARGET_CC) COPT_FLAGS="$(TARGET_CFLAGS)"

$(TARGET_DIR)/usr/sbin/iptables: $(IPTABLES_BUILD_DIR)/iptables
	$(TARGET_CONFIGURE_OPTS) \
	$(MAKE) -C $(IPTABLES_BUILD_DIR) \
		KERNEL_DIR=$(LINUX_DIR) PREFIX=/usr \
		CC=$(TARGET_CC) COPT_FLAGS="$(TARGET_CFLAGS)" \
		DESTDIR=$(TARGET_DIR) install
	$(STRIP) $(TARGET_DIR)/usr/sbin/iptables*
	$(STRIP) $(TARGET_DIR)/usr/lib/iptables/*.so
	rm -rf $(TARGET_DIR)/usr/man

$(TARGET_DIR)/$(IPTABLES_TARGET_INIT_SCRIPT): \
		$(IPTABLES_DIR)/$(IPTABLES_INIT_SCRIPT)
	cp -df $(IPTABLES_DIR)/$(IPTABLES_INIT_SCRIPT) \
		$(TARGET_DIR)/$(IPTABLES_TARGET_INIT_SCRIPT)
	chmod +x $(IPTABLES_DIR)/$(IPTABLES_INIT_SCRIPT)

$(TARGET_DIR)/$(IPTABLES_TARGET_RC_LINK): \
		$(TARGET_DIR)/$(IPTABLES_TARGET_INIT_SCRIPT)
	ln -sf /$(IPTABLES_TARGET_INIT_SCRIPT) $@

iptables: $(TARGET_DIR)/usr/sbin/iptables \
	  $(TARGET_DIR)/$(IPTABLES_TARGET_INIT_SCRIPT) \
	  $(TARGET_DIR)/$(IPTABLES_TARGET_RC_LINK)

iptables-source: $(DL_DIR)/$(IPTABLES_SOURCE)

iptables-clean:
	-$(MAKE) -C $(IPTABLES_BUILD_DIR) clean

iptables-dirclean:
	rm -rf $(IPTABLES_BUILD_DIR)
#############################################################
#
# Toplevel Makefile options
#
#############################################################
ifeq ($(strip $(BR2_PACKAGE_IPTABLES)),y)
TARGETS+=iptables
endif
