#############################################################
#
# libtool
#
#############################################################
LIBTOOL_VER:=1.5.22
LIBTOOL_SOURCE:=libtool-$(LIBTOOL_VER).tar.gz
LIBTOOL_SITE:=http://ftp.gnu.org/pub/gnu/libtool
LIBTOOL_CAT:=zcat
LIBTOOL_DIR:=$(BUILD_DIR)/libtool-$(LIBTOOL_VER)
LIBTOOL_BINARY:=libtool
LIBTOOL_LIBNAME:=libltdl.so.3.1.4
LIBTOOL_TARGET_LIB:=usr/lib/$(LIBTOOL_LIBNAME)
LIBTOOL_TARGET_LIB_LINK:=usr/lib/libltdl.so.3
LIBTOOL_STAGING_LIB:=lib/$(LIBTOOL_LIBNAME)

$(DL_DIR)/$(LIBTOOL_SOURCE):
	 $(WGET) -P $(DL_DIR) $(LIBTOOL_SITE)/$(LIBTOOL_SOURCE)

libtool-source: $(DL_DIR)/$(LIBTOOL_SOURCE)

$(LIBTOOL_DIR)/.unpacked: $(DL_DIR)/$(LIBTOOL_SOURCE)
	$(LIBTOOL_CAT) $(DL_DIR)/$(LIBTOOL_SOURCE) | tar -C $(BUILD_DIR) $(TAR_OPTIONS) -
	touch $(LIBTOOL_DIR)/.unpacked

$(LIBTOOL_DIR)/.configured: $(LIBTOOL_DIR)/.unpacked
	(cd $(LIBTOOL_DIR); rm -rf config.cache; \
		$(TARGET_CONFIGURE_OPTS) \
		CFLAGS="$(TARGET_CFLAGS)" \
		./configure \
		--target=$(GNU_TARGET_NAME) \
		--host=$(GNU_TARGET_NAME) \
		--build=$(GNU_HOST_NAME) \
		--prefix=/usr \
		--exec-prefix=/usr \
		--bindir=/usr/bin \
		--sbindir=/usr/sbin \
		--libexecdir=/usr/lib \
		--sysconfdir=/etc \
		--datadir=/usr/share \
		--localstatedir=/var \
		--mandir=/usr/man \
		--infodir=/usr/info \
		$(DISABLE_NLS) \
	);
	touch $(LIBTOOL_DIR)/.configured

$(LIBTOOL_DIR)/$(LIBTOOL_BINARY): $(LIBTOOL_DIR)/.configured
	$(MAKE) CC=$(TARGET_CC) -C $(LIBTOOL_DIR)
	touch -c $(LIBTOOL_DIR)/$(LIBTOOL_BINARY)

$(STAGING_DIR)/$(LIBTOOL_STAGING_LIB): $(LIBTOOL_DIR)/$(LIBTOOL_BINARY)
	$(MAKE) \
	    prefix=$(STAGING_DIR)/usr \
	    exec_prefix=$(STAGING_DIR)/usr \
	    bindir=$(STAGING_DIR)/usr/bin \
	    sbindir=$(STAGING_DIR)/usr/sbin \
	    libexecdir=$(STAGING_DIR)/usr/lib \
	    datadir=$(STAGING_DIR)/usr/share \
	    sysconfdir=$(STAGING_DIR)/etc \
	    localstatedir=$(STAGING_DIR)/var \
	    libdir=$(STAGING_DIR)/lib \
	    infodir=$(STAGING_DIR)/usr/info \
	    mandir=$(STAGING_DIR)/usr/man \
	    includedir=$(STAGING_DIR)/include \
	    -C $(LIBTOOL_DIR) install;
	$(STRIP) $(STAGING_DIR)/lib/libltdl.so.*.*.* > /dev/null 2>&1
	$(SED) "s,^CC.*,CC=\"/usr/bin/gcc\"," $(STAGING_DIR)/usr/bin/libtool
	$(SED) "s,^LD.*,LD=\"/usr/bin/ld\"," $(STAGING_DIR)/usr/bin/libtool
	rm -rf $(STAGING_DIR)/share/locale $(STAGING_DIR)/usr/info \
		$(STAGING_DIR)/usr/man $(STAGING_DIR)/usr/share/doc

$(TARGET_DIR)/$(LIBTOOL_TARGET_LIB): $(STAGING_DIR)/$(LIBTOOL_STAGING_LIB)
	cp -dpf $(STAGING_DIR)/$(LIBTOOL_STAGING_LIB) \
		$(TARGET_DIR)/$(LIBTOOL_TARGET_LIB)

$(TARGET_DIR)/$(LIBTOOL_TARGET_LIB_LINK): $(TARGET_DIR)/$(LIBTOOL_TARGET_LIB)
	ln -sf `basename $(LIBTOOL_TARGET_LIB)` $@

libtool: $(TARGET_DIR)/$(LIBTOOL_TARGET_LIB) \
	 $(TARGET_DIR)/$(LIBTOOL_TARGET_LIB_LINK)

libtool-clean:
	$(MAKE) DESTDIR=$(TARGET_DIR) CC=$(TARGET_CC) -C $(LIBTOOL_DIR) uninstall
	-$(MAKE) -C $(LIBTOOL_DIR) clean

libtool-cross: $(LIBTOOL_DIR)/$(LIBTOOL_BINARY)

libtool-cross-clean:
	-$(MAKE) -C $(LIBTOOL_DIR) clean

libtool-dirclean:
	rm -rf $(LIBTOOL_DIR)

#############################################################
#
# Toplevel Makefile options
#
#############################################################
ifeq ($(strip $(BR2_PACKAGE_LIBTOOL)),y)
TARGETS+=libtool
endif
