#ifndef _CP_UTIL_H
#define _CP_UTIL_H

#include "common.h"

__BEGIN_DECLS

/**
 * assorted goodies
 */

#ifdef HAVE_CONFIG_H
#include "config.h"

#ifndef HAVE_STRDUP
/**
 * Copies cp_string into newly allocated memory.
 *
 * @retval pointer to copied cp_string
 * @retval NULL if src was NULL
 */
char *strdup(char *src);
#endif /* missing strdup */

#ifndef HAVE_STRNDUP
/**
 * Copies up to maxlen characters of src cp_string into newly allocated memory.
 *
 * @retval pointer to copied cp_string
 * @retval NULL if src was NULL
 */
char *strndup(char *src, int maxlen);
#endif /* missing strndup */

#endif


/**
 * @todo check for buffer overflow
 */
char *str_trim_cpy(char *dst, char *src);


/**
 * scans a cp_string for the first occurence of ch within the first len bytes at
 * most. searching for the null character will return a pointer to the 
 * terminating null. 
 */
char *strnchr(char *str, char ch, int len);


/**
 * Map character 't', 'T' to true.
 */
int parse_boolean(char *value);

/**
 * Map integer value to characters 't', 'f'.
 */
#define format_boolean(val) ((val) ? "t" : "f")


/**
 * Get the current time as long value (UNIX).
 */
long get_timestamp();

/**
 * Return the cp_string or if NULL an empty cp_string.
 */
inline char *dbfmt(char *str);


/**
 * Retrieves the ip-adress of the system where it is running on.
 */
inline unsigned long get_current_ip();

/**
 * Retrieve the ip-address of the system with the 'hostname'.
 */
unsigned long get_host_ip(char *hostname);

/**
 * Formats an ip-address (IPv4) with dot notation.
 */
inline char *ip_to_string(unsigned long ip, char *buf);

/** 
 * converts hex to url encoded binary (eg ABCD => %AB%CD) and returns a newly
 * allocated cp_string
 */
char *hex_url_encode(char *hex);


/** 
 * return a static string coresponding to the posix regcomp error code given
 * as a parameter. the error descriptions come from the gnu man page. 
 */
char *regex_compilation_error(int rc);

/**
 * return a static string describing the stat return code given as a parameter.
 * the resulting error string contains 2 '%s' sequences - the first one could 
 * be used for the program name, function description etc, the second one 
 * for the stat()ed path.
 */
char *stat_error_fmt(int err);

/**
 * check if the directory specified by path exists
 */
int checkdir(char *path);

/**
 * generates a 16 character id based on the current time. The generated id 
 * followed by a terminating null character are written into buf.
 */
void gen_id_str(char *buf);

/**
 * generates a 16 character id based on the given time. The generated id 
 * followed by a terminating null character are written into buf.
 */
void gen_tm_id_str(char *buf, struct timeval *tm);

/* ----------------------------------------------------------------- */

__END_DECLS

/** @} */
#endif

